#include <stdio.h>
#include <stdlib.h>
#include <math.h>

long unsigned int longone, *plug, *halfplug;

int
sizeplug(long unsigned int p)
{
  /* Returns the number of bits 1 in p. */

  long unsigned int p1;
  int i;

  if ( p == 0 ) return(0);
  if ( p == longone ) return(1);

  i = (int)(p&longone);
  p1 = p>>1;

  return(i+sizeplug(p1));
}

int
compareplug(long unsigned int p0, long unsigned int p1)
{
  /* Returns 1 is p0 comes before p1, 0 if equal, -1 if p0 comes after p1.
     Comparison is first by size, second by value. */

  if ( p0 == p1 ) return(0);

  if ( sizeplug(p0) < sizeplug(p1) ) return(1);
  if ( sizeplug(p0) > sizeplug(p1) ) return(-1);

  if ( p0 < p1 ) return(1);
  if ( p0 > p1 ) return(-1);

  return(2); /* You don't exist. Go away. */
}

int
searchplug(long unsigned int p, int i0, int i1)
{
  /* Returns i with plug[i] = p, i0 <= i <= i1.
     We assume plug to ordered as in compareplug.
     Returns -1 is something goes wrong.  */

  extern long unsigned int *plug;

  int i;

  if ( plug[i0] == p ) { return(i0); }
  if ( plug[i1] == p ) { return(i1); }
  if ( i1 - i0 <= 1 ) { return(-1); }
  
  i = (i0+i1)/2;

  if ( plug[i] == p ) { return(i); }

  if ( compareplug(plug[i],p) > 0 ) { return(searchplug(p,i,i1)); }
  else { return(searchplug(p,i0,i)); }

  /* return(-1); */
}

int
main(int argc, char **argv)
{ 
  extern long unsigned int longone, *plug;
  long unsigned int p, pp, Nlong;
  int **adj, **adjk, *cp, *cpt, i, ii, j, jj, jjx, k, kk, n, nmax, N, NN, s;
  int *hbl, *hwh;
  int **pm, **pmt;
  FILE *pin, *pout, *poutt;

  if ( argc != 4 )
  {
    fprintf(stderr,"Usage: ./adj infile.txt outfile.txt outfilet.txt.\n");
    fprintf(stderr,"The file infile.txt should contain the K matrix.\n");
    fprintf(stderr,"The files outfile.txt and outfilet.txt are overwritten.\n");
    return(1);
  }


  longone = 1;
  nmax = (int)(4*sizeof(long));

  pin = fopen(argv[1],"r");

  fscanf(pin,"%d",&n); 


  if ( n > nmax )
  {
    fprintf(stderr,"n is too large. Largest value is %d.\n\n",nmax);
    return(1);
  }

  hbl = (int *) malloc(n*sizeof(int));
  hwh = (int *) malloc(n*sizeof(int));

  adj = (int **) malloc(n*sizeof(int *));
  adjk = (int **) malloc(n*sizeof(int *));
  for ( i = 0 ; i < n ; i++ )
  {
    adj[i] = (int *) malloc(n*sizeof(int));
    adjk[i] = (int *) malloc(n*sizeof(int));
  }

  for ( i = 0 ; i < n ; i++ ) for ( j = 0 ; j < n ; j++ )
  {
    fscanf(pin,"%d",adjk[i]+j); 
    adj[i][j] = (adjk[i][j])*(adjk[i][j]);
  }

  fclose(pin);


  NN = 1<<n;
  halfplug = (long unsigned int *) malloc(NN*sizeof(long unsigned int));

  k = 0;
  for ( j = 0 ; j <= n ; j++ )
  for ( i = 0 ; i < NN ; i++ )
  if ( sizeplug(i) == j )
  {
    halfplug[k] = i;
    k++;
  }
  
  if ( k != NN ) 
  {
    fprintf(stderr,"Error generating halfplug. \n\n");
    return(1);
  }


  Nlong = 1; j = 1; 
  for ( i = 0 ; i < n ; i++ ) { Nlong *= n+i+1; j *= i+1; }
  N = Nlong/j;

  fprintf(stdout,"%d %d %d\n",n,NN,N); 

  cp = (int *) malloc(N*sizeof(int));
  cpt = (int *) malloc(N*sizeof(int));

  /* fprintf(stdout,"cpp allocated.\n\n"); */

  plug = (long unsigned int *) malloc(N*sizeof(long unsigned int));

  /* fprintf(stdout,"plug allocated.\n\n"); */

  for ( k = 0 ; k < N ; k++ ) { plug[k] = 0; cp[k] = 0; cpt[k] = 0; }

  /* fprintf(stdout,"plug cleaned.\n\n"); */

  plug[0] = 0; cp[0] = 1; cpt[0] = 1;

  k = 1;

  for ( i = 1 ; i < NN ; i++ )
  for ( j = 1 ; j < NN ; j++ )
  if ( sizeplug(halfplug[j]) == sizeplug(halfplug[i]) )
  {
    if ( k >= N )
    {
      fprintf(stderr,"Error generating plug. (%d,%d) \n\n",i,j);
      return(1);
    }

    p = (halfplug[i]<<n) + halfplug[j];
    plug[k] = p;

    cp[k] = 0;
    cpt[k] = 0;
    
    ii = 0;
    while ( ( halfplug[i]&(longone<<ii) ) == 0 ) ii++;

    jjx = 1;
    for ( jj = 0 ; jj < n ; jj++ )
    if ( ( halfplug[j]&(longone<<jj) ) > 0 )
    {
      if ( adj[ii][jj] > 0 )
      {
        pp = p - (longone<<(n+ii)) - (longone<<jj);
        kk = searchplug(pp,0,k);
        if ( kk < 0 )
        {
          fprintf(stderr,"Searchplug error. (%lX) \n\n",pp);
          return(1);
        }
        cp[k] += adj[ii][jj]*cp[kk];
        cpt[k] += jjx*adjk[ii][jj]*cpt[kk];
      }

      jjx *= -1;
    }


    k++;
  }


  if ( k != N )
  {
     fprintf(stderr,"Error generating plug. (END) \n\n");
     return(1);
  }

  /*
  for ( k = 0 ; k < N ; k++ )
  fprintf(stdout,"%d %lX %d\n",k,plug[k],cp[k]);
  */

  pm = (int **) malloc(N*sizeof(int *));
  pmt = (int **) malloc(N*sizeof(int *));
  for ( i = 0 ; i < N ; i++ )
  {
    pm[i] = (int *) malloc(N*sizeof(int));
    pmt[i] = (int *) malloc(N*sizeof(int));
  }

  for ( i = 0 ; i < N ; i++ )
  for ( j = 0 ; j < N ; j++ )
  {
    if ( (plug[i]&plug[j]) > 0 )
    { pm[i][j] = 0; pmt[i][j] = 0; }
    else
    { 
      p = plug[N-1]^(plug[i]^plug[j]);
      k = searchplug(p,0,N-1);
      if ( k < 0 )
      {
        fprintf(stderr,"Searchplug error. (%lX) \n\n",p);
        return(1);
      }
      pm[i][j] = cp[k];

      for ( ii = 0 ; ii < n ; ii++ )
      {
        hbl[ii] = 0;
        hwh[ii] = 0;
        if ( ((1<<ii)&(plug[i])) > 0 ) hbl[ii] = 1;
        if ( ((1<<ii)&(plug[j])) > 0 ) hbl[ii] = -1;
        if ( ((1<<(n+ii))&(plug[i])) > 0 ) hwh[ii] = 1;
        if ( ((1<<(n+ii))&(plug[j])) > 0 ) hwh[ii] = -1;
      }

      s = 1;

      for ( ii = 0 ; ii < n ; ii++ )
      for ( jj = ii+1 ; jj < n ; jj++ )
      {
        if ( hbl[ii] > hbl[jj] ) s *= -1;
        if ( hwh[ii] > hwh[jj] ) s *= -1;
      }

      pmt[i][j] = s*cpt[k];
    }
  }

  pout = fopen(argv[2],"w");
  poutt = fopen(argv[3],"w");

  fprintf(pout,"%d\n",N);
  fprintf(poutt,"%d\n",N);
  for ( i = 0 ; i < N ; i++ )
  for ( j = 0 ; j < N ; j++ )
  {
    fprintf(pout,"%d %d %d\n",i,j,pm[i][j]);
    fprintf(poutt,"%d %d %d\n",i,j,pmt[i][j]);
  }

  fclose(pout);
  fclose(poutt);

  /*

  fprintf(stdout,"aaa := Matrix(1..%d,1..%d):\n\n",N,N);
  fprintf(stdout,"aat := Matrix(1..%d,1..%d):\n\n",N,N);
  for ( i = 0 ; i < N ; i++ )
  {
    for ( j = 0 ; j < N ; j++ )
    {
      fprintf(stdout,"aaa[%d,%d] := %d:\n",i+1,j+1,pm[i][j]);
      fprintf(stdout,"aat[%d,%d] := %d:\n",i+1,j+1,pmt[i][j]);
    }
    fprintf(stdout,"\n");
  }




  for ( i = 0 ; i < n ; i++ )
  {
    fprintf(stdout,"\n");
    for ( j = 0 ; j < n ; j++ ) fprintf(stdout,"%d",adj[i][j]);
  }
  fprintf(stdout,"\n\n");
  */

  return(0);
}


  

