#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define N 75
/* N is an upper bound on the total number of cubes */

char **ltilings;
int *ltwists;
long int *lcomps;
long int NT, n, np, nm, NT0, NT1;
int NN, Nh; /* Correct value for number of cubes and its half */
int color[N], pospt[N], negpt[N], clabelpt[N];
int point[8][N];
int KK[N][N];

/* A tiling is encoded by a string of letters from 'a' to 'h';
   '0' denotes an empty cell (for partial tilings).
   'a' means the companion cube is obtained by increasing coordinate 1 by 1.
   'b' means the companion cube is obtained by decreasing coordinate 1 by 1.
   'c' means the companion cube is obtained by increasing coordinate 2 by 1.
   'd' means the companion cube is obtained by decreasing coordinate 2 by 1.
*/

int
dist_tiling(char tiling0[], char tiling1[])
{
  /* Computes and returns the distance d between tiling0 and tiling1.
     This is always an even non negative integer;
     d = 0 if and only if they are equal,
     d = 4 if and only if they differ by a flip,
     d > 4 otherwise. */

  int d, i;

  d = 0;
  for ( i = 0 ; i < NN ; i++ ) if ( tiling0[i] != tiling1[i] ) { d++; }
  return(d);
}

int
compare_tiling(char tiling0[], char tiling1[])
{
  /* Returns 1 if tiling0 comes before tiling1;
            -1 if after; 0 if equal. */
  int i;

  for ( i = 0 ; i < NN ; i++ )
  {
    if ( tiling0[i] < tiling1[i] ) { return(1); }
    if ( tiling0[i] > tiling1[i] ) { return(-1); }
  }

  return(0);
}

long int
num_tiling(char tiling[], long int nmin, long int nmax)
{
  long int nmed, nansw;
  int c;

  if ( nmax - nmin < 2 )
  {
    c = compare_tiling(tiling,ltilings[nmin]);
    if ( c == 0 ) { return(nmin); }
    c = compare_tiling(tiling,ltilings[nmax]);
    if ( c == 0 ) { return(nmax); }
    return((long)(-1));
  }
  else
  {
    nmed = (nmin+nmax)/2;
    c = compare_tiling(tiling,ltilings[nmed]);

    if ( c == 0 ) { return(nmed); }

    if ( c > 0 ) { nansw = num_tiling(tiling,nmin,nmed); }
    else if ( c < 0 ) { nansw = num_tiling(tiling,nmed,nmax); }

    if ( nansw < 0 ) return(nansw);

    c = compare_tiling(tiling,ltilings[nansw]);

    if ( c == 0 ) { return(nansw); }
    
    return((long)(-1));
  }
}

int
flip_tiling(int i0, int dir, char tiling0[], char tiling1[])
{
  /* tiling0 is input, tiling1 is output.
     tiling1 is returned either as a flip of tiling0
     at position i0 and direction dir or,
     if that is not possible, as a copy of tiling0.
     Returns 1 if tiling is possible, 0 otherwise.

     extern long int n;
  */

  extern int point[8][N];

  int i, i1, i2, i3, k;

  i1 = point[dir][i0];
  if ( ( i1 < 0 ) || ( tiling0[i1] != tiling0[i0] ) ) return(0);

  for ( i = 0 ; i <= NN ; i++ ) { tiling1[i] = tiling0[i]; }

  /* sprintf(tiling1,"%s",tiling0); */

 
  k = tiling0[i0] - 'a';
  i2 = point[k^1][i0];
  i3 = point[k^1][i1];

  tiling1[i0] = 'a'+(dir^1);
  tiling1[i1] = 'a'+dir;
  tiling1[i2] = 'a'+(dir^1);
  tiling1[i3] = 'a'+dir;

/*
  n1 = num_tiling(tiling1,(long)(0),n-1);

  if ( n1 < 0 )
  { 
    fprintf(stderr,"BAD FLIP!");
  }
*/

  return(1);
}


int 
read_region(FILE *pin)
{
  extern int NN;
  extern int color[], pospt[], negpt[], clabelpt[];
  extern int point[8][N];
  extern int KK[N][N];

  char checkstr[256];
  int i, ip, in, dir;

  fscanf(pin,"%d",&NN);
  if ( (NN&1) == 1 ) return(-2);
  Nh = NN/2;

  for ( i = NN ; i < N ; i++ )
  {
    color[i] = 0;
    clabelpt[i] = -1;
    for ( dir = 0 ; dir < 8 ; dir++ ) point[dir][i] = -1;
  }

  ip = 0; in = 0;
  fscanf(pin,"%s",checkstr);
  for ( i = 0 ; i < NN ; i++ )
  {
    if ( checkstr[i] == '+' )
    { color[i] = 1; clabelpt[i] = ip; pospt[ip] = i; ip++; }
    else if ( checkstr[i] == '-' )
    { color[i] = -1; clabelpt[i] = in; negpt[in] = i; in++; }
    else { return(-4); }
  }
  if ( ( ip != Nh ) || ( in != Nh ) ) return(-5);
  for ( i = Nh ; i < N ; i++ ) { pospt[i] = -1; negpt[i] = -1; }

  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'a' ) return(-8);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[0]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'b' ) return(-9);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[1]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'c' ) return(-10);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[2]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'd' ) return(-11);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[3]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'e' ) return(-12);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[4]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'f' ) return(-13);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[5]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'g' ) return(-14);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[6]+i);
  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'h' ) return(-15);
  for ( i = 0 ; i < N ; i++ ) fscanf(pin,"%d",point[7]+i);

  fscanf(pin,"%s",checkstr);
  if ( checkstr[0] != 'K' ) return(-50);

  for ( ip = 0 ; ip < Nh ; ip++ )
  for ( in = 0 ; in < Nh ; in++ )
  {
    fscanf(pin,"%d",KK[ip]+in);
  }


  fscanf(pin,"%s",checkstr);

  if ( ( checkstr[0] != 'E' ) || ( checkstr[1] != 'N' ) || ( checkstr[2] != 'D' ) )
  { return(-1); }

  return(NN);
}


int
sign_perm(int perm[])
{
  /* Computes and returns the sign of a permutation of size Nh. */

  extern int Nh;

  int i0, i1, sign;

  sign = 1;
  for ( i0 = 0 ; i0 < Nh ; i0++ )
  for ( i1 = i0+1 ; i1 < Nh ; i1++ )
  if ( perm[i0] > perm[i1] ) { sign *= -1; }

  return(sign);
}

int
sign_perm_tiling(char tiling[], int perm[])
{
  /* Constructs the permutation perm corresponding to a tiling.
     Returns the sign of perm[]. 
  */

  extern int Nh, point[8][N], pospt[], clabelpt[];
  int  i, ii, k;
  char c;

  for ( i = 0 ; i < Nh ; i++ )
  {
    ii = pospt[i];
    c = tiling[ii];
    k = c - 'a';
    if ( ( k >= 0 ) && ( k < 8 ) ) { perm[i] = clabelpt[point[k^1][ii]]; }
    else perm[i] = -1;
  }

  return(sign_perm(perm));
}

int
twist_tiling(char tiling[])
{
  /* Computes the twist of a tiling with values in {+1,-1}. */

  extern int KK[N][N];

  int perm[N], Ks, sp, i0;

  sp = sign_perm_tiling(tiling, perm);

  Ks = 1;
  for ( i0 = 0 ; i0 < Nh ; i0++ ) { Ks *= KK[i0][perm[i0]]; }

  return(sp*Ks);
}



int
fill_step(FILE *pout, char *intiling)
{
  extern long int n, np, nm;
  extern int point[8][N];

  char outtiling[N+2];
  int i, k, twist;
 
  for ( i = 0 ; ( ( intiling[i] != '0' ) && ( intiling[i] != '\0' ) ) ; i++ )
  {}

  if ( ( i == NN ) || ( intiling[i] == '\0' ) )
  {
    intiling[NN] = '\0';
    twist = twist_tiling(intiling);
    fprintf(pout,"%ld %s %d\n",n,intiling,twist);
    if ( n < NT )
    {
      sprintf(ltilings[n],"%s",intiling);
      ltilings[n][NN] = '\0';
      ltwists[n] = twist;
      lcomps[n] = n;
    }

    n++;
    if ( twist > 0 ) { np++; }
    else if ( twist < 0 ) { nm++; }
    return(0);
  }

  for ( k = 0 ; k < 8 ; k++ )
  if ( ( point[k][i] > i ) && ( intiling[point[k][i]] == '0' ) )
  {
    sprintf(outtiling,"%s",intiling);
    outtiling[NN] = '\0';
    outtiling[point[k][i]] = 'a'+k;
    outtiling[i] = 'a'+(k^1);
    fill_step(pout,outtiling);
  }

  return(1);
}

int
main(int argc, char **argv)
{ 
  extern char **ltilings;
  extern int *ltwists;
  extern long int n, np, nm;
  extern int NN;

  long int nt, n0, n1, nc, nd, nx;
  int i, dir;
  char xtiling[N+2];
  FILE *pin, *pout;

  n = 0;
  np = 0;
  nm = 0;

  if ( argc != 8 )
  {
    fprintf(stderr,"Usage: ./adj infile0.txt infile1.txt outfile.txt NN NT NT0 NT1.\n");
    fprintf(stderr,"The file infile0.txt should contain a region.\n");
    fprintf(stderr,"The file infile1.txt should contain a list of tilings.\n");
    fprintf(stderr,"The file outfile.txt is overwritten with a list of tilings.\n");
    fprintf(stderr,"NT is the exact number of tilings. \n");
    fprintf(stderr,"Components up to N0 should have been computed already. \n");
    fprintf(stderr,"Components up to N1 will be computed and written. \n");
    return(1);
  }

  NN = atoi(argv[4]);
  NT = atol(argv[5]);
  NT0 = atol(argv[6]);
  NT1 = atol(argv[7]);

  if ( NT1 > NT ) NT1 = NT;

  pin = fopen(argv[1],"r");
  i = read_region(pin);
  fclose(pin);

  if ( i < 0 )
  {
    fprintf(stderr,"Infile corrupted: %d\n",i);
    return(1);
  }


  ltilings = (char **) malloc(NT*sizeof(char*));
  ltwists = (int *) malloc(NT*sizeof(int));
  lcomps = (long *) malloc(NT*sizeof(long));
  for ( nt = 0 ; nt < NT ; nt++ )
  { ltilings[nt] = (char *) malloc((NN+1)*sizeof(char)); }

  n = NT; /* For consistency with old code. */

  pin = fopen(argv[2],"r");
  for ( n0 = 0 ; n0 < NT ; n0++ )
  {
    fscanf(pin,"%ld %s %d %ld",&nt,ltilings[n0],ltwists+n0,lcomps+n0);
    if ( ( nt != n0 ) || ( lcomps[n0] > n0 ) )
    {
      fprintf(stderr,"Infile corrupted!\n\n");
      return(1);
    }
  } 
  fclose(pin);
 

/* Fast method. */


  for ( n1 = NT0 ; n1 < NT1 ; n1++ )
  {
  for ( i = 0 ; i < NN ; i++ )
  for ( dir = 0 ; dir < 8 ; dir++ )
  if ( flip_tiling(i,dir,ltilings[n1],xtiling) > 0 )
  if ( compare_tiling(xtiling,ltilings[n1]) > 0 )
  {
    n0 = num_tiling(xtiling,(long)(0),n-1);
    if ( n0 < 0 )
    {
      fprintf(stderr,"%ld %d %d %s %s\n",n1,i,dir,ltilings[n1],xtiling);
    }
    if ( ( n0 < n1 ) && ( lcomps[n0] != lcomps[n1] ) )
    {
      if ( lcomps[n1] == n1 )
      {
        if ( lcomps[n0] > n1 )
        { fprintf(stderr,"\n%ld %ld %ld %ld\n",n1,n0,lcomps[n1],lcomps[n0]); }
        lcomps[n1] = lcomps[n0];
      }
      else
      {
        if ( lcomps[n0] < lcomps[n1] ) { nc = lcomps[n0]; nd = lcomps[n1]; }
        else { nc = lcomps[n1]; nd = lcomps[n0]; }
        for ( nx = 0 ; nx <= n1 ; nx++ )
        if ( lcomps[nx] == nd ) { lcomps[nx] = nc; }
      }
    }
  }
  }


  pout = fopen(argv[3],"w");

  for ( n0 = 0 ; n0 < NT ; n0++ )
  {
    fprintf(pin,"%ld %s %d %ld\n",n0,ltilings[n0],ltwists[n0],lcomps[n0]);
  } 

  if ( NT1 == NT )
  {
    fprintf(pout,"\n\nSIZES OF COMPONENTS\n\n");
    for ( n0 = 0 ; n0 < n ; n0++ )
    if ( lcomps[n0] == n0 )
    {
      nx = 0;
      for ( n1 = n0 ; n1 < n ; n1++ )
      if ( lcomps[n1] == n0 ) { nx++; }
      fprintf(pout,"%ld %ld %d\n",n0,nx,ltwists[n0]); 
    }
  }

  fclose(pout);


  return(0);
}


  

